﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Collections.ObjectModel;

namespace Ellab.Charting
{
    public partial class Chart : UserControl
    {
        private const string HORIZONTAL_AXIS_NAME = "HORIZONTAL_AXIS";
        private Steema.TeeChart.Animations.Expand expand = null;

        internal Steema.TeeChart.TChart TChart { get { return chartControl; } }

        #region Events

        public new event MouseEventHandler MouseUp;
        private void TChart_MouseUp(object sender, MouseEventArgs e)
        {
            if (MouseUp != null)
                MouseUp(this, e);
        }
        public new event MouseEventHandler MouseDown;
        private void TChart_MouseDown(object sender, MouseEventArgs e)
        {
            if (MouseDown != null)
                MouseDown(this, e);
        }

        #endregion Events

        public Chart()
        {
            InitializeComponent();

            TChart.Aspect.View3D = false;
            TChart.Chart.Panel.MarginUnits = Steema.TeeChart.PanelMarginUnits.Pixels;
            TChart.Chart.Panel.MarginBottom = 50;
            TChart.Chart.Panel.MarginLeft = 0;
            TChart.Chart.Panel.MarginRight = 10;

            TChart.Chart.Header.Alignment = StringAlignment.Far;
            chartControl.Header.Color = System.Drawing.Color.Black;
            Title = "";

            Zoom.ZoomInHandler = Zoom.DefaultZoomInHandler;
            Zoom.ZoomOutHandler = Zoom.DefaultZoomOutHandler;

            TChart.Aspect.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;

            TChart.BeforeDraw += new Steema.TeeChart.PaintChartEventHandler(RedrawingHandler);
            TChart.AfterDraw += new Steema.TeeChart.PaintChartEventHandler(RedrawnHandler);

            TChart.MouseUp += new MouseEventHandler(TChart_MouseUp);
            TChart.MouseDown += new MouseEventHandler(TChart_MouseDown);
        }

        public void EnablePointExpansion()
        {
            expand = new Steema.TeeChart.Animations.Expand();
            chartControl.Animations.Add(expand);
            expand.Target = Steema.TeeChart.ChartClickedPartStyle.SeriesPointer;
            expand.Trigger = Steema.TeeChart.Animations.AnimationTrigger.MouseOver;
            expand.SizeBy = 5;
            expand.Speed = 20;
        }

        public Rectangle ChartBounds
        {
            get
            {
                return new Rectangle
                (
                    Convert.ToInt32(TChart.Chart.Panel.MarginLeft),
                    Convert.ToInt32(TChart.Chart.Panel.MarginTop),
                    this.Width - Convert.ToInt32(TChart.Chart.Panel.MarginRight),
                    //this.Height - HorizontalAxis.TAxis.Position
                    this.Height - Convert.ToInt32(TChart.Chart.Panel.MarginBottom)
                );
            }
        }

        #region Chart Export

        public class ChartExport
        {
            private Chart parent = null;
            internal ChartExport(Chart _parent)
            {
                parent = _parent;

                Width = 800;
                Height = 600;
                Format = ExportFormat.BMP;
            }

            public Size Size
            {
                get { return new Size(Width, Height); }
                set { Width = value.Width; Height = value.Height; }
            }
            public int Width { get; set; }
            public int Height { get; set; }
            public ExportFormat Format { get; set; }

            public void ToClipboard()
            {
                Steema.TeeChart.Export.ImageExportFormat tExport = GetTChartFormat(Format);
                tExport.Width = Width;
                tExport.Height = Height;
                tExport.CopyToClipboard();
            }

            public void ToFile(string filename)
            {
                Steema.TeeChart.Export.ImageExportFormat tExport = GetTChartFormat(Format);
                tExport.Width = Width;
                tExport.Height = Height;
                tExport.Save(filename);
            }

            public void ToStream(System.IO.Stream stream)
            {
                Steema.TeeChart.Export.ImageExportFormat tExport = GetTChartFormat(Format);
                tExport.Width = Width;
                tExport.Height = Height;
                tExport.Save(stream);
            }

            public byte[] ToArray()
            {
                System.IO.MemoryStream stream = new System.IO.MemoryStream();
                ToStream(stream);
                return stream.ToArray();
            }

            public enum ExportFormat { BMP, JPEG, PNG } //Remember to modify GetTChartFormat when adding to this enum!

            private Steema.TeeChart.Export.ImageExportFormat GetTChartFormat(ExportFormat format)
            {
                switch (format)
                {
                    case ExportFormat.JPEG: return parent.TChart.Export.Image.JPEG;
                    case ExportFormat.PNG: return parent.TChart.Export.Image.PNG;
                    default: return parent.TChart.Export.Image.Bitmap;  //ExportFormat.BMP handled here
                }
            }
        }
        private ChartExport export = null;
        public ChartExport Export
        {
            get
            {
                if (export == null)
                    export = new ChartExport(this);
                return export;
            }
        }

        #endregion Chart Export

        #region Updating/Redrawing

        public bool SuspendUpdates
        {
            get { return !TChart.AutoRepaint; }
            set
            {
                if (suspendUpdatesDelegate == null)
                    suspendUpdatesDelegate = new Action<bool>(SetSuspendUpdates);
                InvokeDelegate(suspendUpdatesDelegate, value);
            }
        }
        private Action<bool> suspendUpdatesDelegate = null;
        private void SetSuspendUpdates(bool value)
        {
            TChart.AutoRepaint = !value;
            if (value == false)
            {
                ForceUpdate();
            }
        }

        public void ForceUpdate()
        {
            if (forceUpdateDelegate == null)
                forceUpdateDelegate = new Action(ForceUpdateInternal);
            InvokeDelegate(forceUpdateDelegate);
        }
        private Action forceUpdateDelegate = null;
        private void ForceUpdateInternal()
        {
            TChart.Invalidate();
        }

        public event EventHandler Redrawing;
        private void RedrawingHandler(object sender, Steema.TeeChart.Drawing.Graphics3D g)
        {
            if (Redrawing != null)
                Redrawing(this, new EventArgs());
        }
        public event EventHandler Redrawn;
        private void RedrawnHandler(object sender, Steema.TeeChart.Drawing.Graphics3D g)
        {
            if (Redrawn != null)
                Redrawn(this, new EventArgs());
        }

        #endregion

        #region Aesthetics

        public string Title
        {
            get { return chartControl.Header.Text; }
            set
            {
                if (titleDelegate == null)
                    titleDelegate = new Action<string>(SetTitle);
                InvokeDelegate(titleDelegate, value);
            }
        }
        private Action<string> titleDelegate = null;
        private void SetTitle(string value)
        {
            chartControl.Header.Text = value;
        }

        public bool LegendVisible
        {
            get { return chartControl.Legend.Visible; }
            set
            {
                if (legendVisibleDelegate == null)
                    legendVisibleDelegate = new Action<bool>(SetLegendVisible);
                InvokeDelegate(legendVisibleDelegate, value);
            }
        }
        private Action<bool> legendVisibleDelegate = null;
        private void SetLegendVisible(bool value)
        {
            chartControl.Legend.Visible = value;
        }

        public string ToolTip
        {
            set
            {
                if (toolTipDelegate == null)
                    toolTipDelegate = new Action<string>(SetToolTip);
                InvokeDelegate(toolTipDelegate, value);
            }
        }
        private Action<string> toolTipDelegate = null;
        private void SetToolTip(string value)
        {
            if (value != null && value != "")
            {
                TChart.Chart.ToolTip.Text = value;
                TChart.Chart.ToolTip.Show();
                TChart.Chart.ToolTip.InitialDelay = 1500;
            }
            else
            {
                TChart.Chart.ToolTip.Hide();
            }
        }



        #endregion

        #region Axes

        private Axis horizontalAxis = null;
        public Axis HorizontalAxis
        {
            get
            {
                if (horizontalAxis == null)
                {
                    horizontalAxis = new Axis(this, "Time", true);
                    TChart.Axes.Custom.Add(horizontalAxis.TAxis);
                }
                return horizontalAxis;
            }
        }

        private AxisCollection axes = new AxisCollection();
        public AxisCollection VerticalAxes
        {
            get { return axes; }
        }

        public Axis GetAxis(string unit)
        {
            if (!axes.Contains(unit) && unit != HORIZONTAL_AXIS_NAME)
            {
                axes.Add(new Axis(this, unit));

                if (addAxisDelegate == null)
                    addAxisDelegate = new Action<Steema.TeeChart.Axis>(AddAxisInternal);
                InvokeDelegate(addAxisDelegate, axes[unit.ToUpper()].TAxis);
            }
            return axes[unit.ToUpper()];
        }
        private Action<Steema.TeeChart.Axis> addAxisDelegate = null;
        private void AddAxisInternal(Steema.TeeChart.Axis value)
        {
            TChart.Axes.Custom.Add(value);

            RepositionAxes();
        }
        private void RepositionAxes()
        {
            TChart.Chart.Panel.MarginLeft = axes.Sum(a => a.Width);
            int i = 0;
            foreach (Axis a in axes)
            {
                a.Reposition(i);
                i++;
            }
        }
        internal void RemoveUnusedAxes()
        {
            HashSet<string> usedAxes = new HashSet<string>();
            foreach (ISeries s in Series)
            {
                usedAxes.Add(s.Unit);
            }
            HashSet<string> unusedAxes = new HashSet<string>();
            foreach (Axis a in VerticalAxes)
            {
                if (!usedAxes.Contains(a.Unit))
                    unusedAxes.Add(a.Unit);
            }
            foreach (string a in unusedAxes)
            {
                RemoveAxis(a);
            }
            RepositionAxes();
        }
        private void RemoveAxis(string unit)
        {
            if (axes.Contains(unit))
            {
                axes[unit].Remove();
                axes.Remove(unit);
            }
        }
        private bool IsAxisUsed(string unit)
        {
            int count = (from s in Series
                         where s.Unit.ToUpper() == unit.ToUpper()
                         select s).Count();
            return count > 1; //Return true if more than 0 series are using axis
        }

        public int AxisOffset
        {
            set
            {
                foreach (Axis axis in axes)
                {
                    axis.Offset = value;
                }
            }
        }

        #endregion

        #region Series

        public event EventHandler<SeriesEventArgs> SeriesAdded;
        public event EventHandler<SeriesEventArgs> SeriesRemoved;

        private SeriesCollection series = new SeriesCollection();
        public SeriesCollection Series
        {
            get
            {
                return series;
            }
        }

        public ISeries CreateSeries(SeriesType type, string unit)
        {
            int i = 1;
            while (series.Contains("Series " + i))
                i++;
            return CreateSeries(type, "Series " + i, unit);
        }
        public ISeries CreateSeries(SeriesType type, string name, string unit)
        {
            ISeries s = SeriesFactory.CreateSeries(type, this, name, unit);

            return s;
        }

        internal void AddSeries(ISeries s)
        {
            if (series.Contains(s.Name))
                throw new Exception("Series with name \"" + s.Name + "\" already exists.");

            s.Visible = true;
            s.XValuesAsDataTime = true;

            if (series != null)
            {
                series.Add(s);
            }

            if (SeriesAdded != null)
                SeriesAdded(this, new SeriesEventArgs(s));
        }

        public void RemoveSeries(string name)
        {
            if (series.Contains(name))
            {
                ISeries s = series[name];

                series[name].Remove();
            }
        }

        /// <summary>
        /// For internal use only. This is called by ISeries' Remove method and should never be called from anywhere else!
        /// </summary>
        /// <param name="s"></param>
        internal void RemoveSeries(ISeries s)
        {
            if (series.Contains(s.Name))
            {
                if (!IsAxisUsed(s.Unit))
                {
                    RemoveAxis(s.Unit);
                    RepositionAxes();
                }

                series.Remove(s.Name);

                if (SeriesRemoved != null)
                    SeriesRemoved(this, new SeriesEventArgs(s));
            }
        }

        public void RemoveAllSeries()
        {
            try
            {
                foreach (ISeries s in series)
                {
                    s.RemoveAll();
                }
                series.Clear();
            }
            catch (Exception ex)
            {
                Ellab.Debug.GlobalDebug.debugForm.WriteText(ex);
            }
        }

        #endregion

        #region Tools

        public event EventHandler<ToolEventArgs> ToolAdded;
        public event EventHandler<ToolEventArgs> ToolRemoved;

        private ToolCollection tools = new ToolCollection();
        public ToolCollection Tools
        {
            get
            {
                return tools;
            }
        }

        public ITool CreateTool(ToolType type)
        {
            int i = 1;
            while (tools.Contains("Tool " + i))
                i++;
            return CreateTool(type, "Tool " + i);
        }
        public ITool CreateTool(ToolType type, string name)
        {
            return ToolFactory.CreateTool(type, this, name);
        }

        internal void AddTool(ITool t)
        {
            if (tools.Contains(t.Name))
                throw new Exception("Tool with name \"" + t.Name + "\" already exists.");

            if (tools != null)
            {
                tools.Add(t);

                if (ToolAdded != null)
                    ToolAdded(this, new ToolEventArgs(t));
            }
        }

        public void RemoveTool(string name)
        {
            if (tools.Contains(name))
            {
                ITool t = tools[name];

                t.Remove();
            }
        }

        internal void RemoveTool(ITool t)
        {
            if (tools.Contains(t.Name))
            {
                tools.Remove(t.Name);

                if (ToolRemoved != null)
                    ToolRemoved(this, new ToolEventArgs(t));
            }
        }

        public void RemoveAllTools()
        {
            foreach (ITool t in tools)
            {
                t.Remove();
            }
        }

        #endregion

        #region Zoom

        public class ChartZoom
        {
            private Chart parent = null;
            internal ChartZoom(Chart _parent)
            {
                parent = _parent;

                parent.TChart.Zoomed += delegate(object sender, EventArgs e)
                {
                    if (ZoomInHandler != null)
                    {
                        ZoomInHandler(sender, e);
                    }
                };

                DefaultZoomInHandler = new EventHandler(ZoomIn);
                CustomDefaultZoomInHandler = new EventHandler(CustomZoomIn);

                parent.TChart.UndoneZoom += delegate(object sender, EventArgs e)
                {
                    if (ZoomOutHandler != null)
                        ZoomOutHandler(sender, e);
                };
                DefaultZoomOutHandler = new EventHandler(ZoomOut);
            }

            public bool Enable
            {
                get { return parent.TChart.Zoom.Allow; }
                set
                {
                    if (enableDelegate == null)
                        enableDelegate = new Action<bool>(SetEnable);
                    parent.InvokeDelegate(enableDelegate, value);
                }
            }
            private Action<bool> enableDelegate = null;
            private void SetEnable(bool value)
            {
                parent.TChart.Zoom.Allow = value;
            }

            public EventHandler ZoomInHandler { get; set; }
            public EventHandler DefaultZoomInHandler { get; private set; }
            public EventHandler CustomZoomInHandler { get; set; }
            public EventHandler CustomDefaultZoomInHandler { get; private set; }
            public EventHandler ZoomOutHandler { get; set; }
            public EventHandler DefaultZoomOutHandler { get; private set; }

            private Stack<Dictionary<string, AxisZoomState>> zoomHistory = new Stack<Dictionary<string, AxisZoomState>>();

            private void ZoomIn(object sender, EventArgs e)
            {
                if (zoomInDelegate == null)
                {
                    zoomInDelegate = new Action<object, EventArgs>(ZoomInInternal);
                }

                parent.InvokeDelegate(zoomInDelegate, sender, e);
            }
            private Action<object, EventArgs> zoomInDelegate = null;
            private void ZoomInInternal(object sender, EventArgs e)
            {
                Dictionary<string, AxisZoomState> state = new Dictionary<string, AxisZoomState>();
                foreach (Axis axis in parent.axes)
                {
                    state[axis.Unit] = new AxisZoomState()
                    {
                        AutomaticMinimum = parent.axes[axis.Unit].AutomaticMinimum,
                        AutomaticMaximum = parent.axes[axis.Unit].AutomaticMaximum,
                        Minimum = parent.axes[axis.Unit].Minimum,
                        Maximum = parent.axes[axis.Unit].Maximum
                    };
                }
                state[HORIZONTAL_AXIS_NAME] = new AxisZoomState()
                {
                    AutomaticMinimum = parent.horizontalAxis.AutomaticMinimum,
                    AutomaticMaximum = parent.horizontalAxis.AutomaticMaximum,
                    Minimum = parent.horizontalAxis.Minimum,
                    Maximum = parent.horizontalAxis.Maximum
                };
                zoomHistory.Push(state);

                foreach (Axis a in parent.axes)
                {
                    a.Minimum = a.PixelCoordinateToValue(parent.TChart.Zoom.x1, parent.TChart.Zoom.y1);
                    a.Maximum = a.PixelCoordinateToValue(parent.TChart.Zoom.x0, parent.TChart.Zoom.y0);
                }
                parent.horizontalAxis.Minimum = parent.horizontalAxis.PixelCoordinateToValue(parent.TChart.Zoom.x1, parent.TChart.Zoom.y1);
                parent.horizontalAxis.Maximum = parent.horizontalAxis.PixelCoordinateToValue(parent.TChart.Zoom.x0, parent.TChart.Zoom.y0);
            }

            #region Custom Zooming
            /// <summary>
            /// Class used to hold axis (horizontal as well as vertical) values, corresponding to the user defined zoom rectangle.
            /// For obvious reasons, if these zoom coordinates are saved as pixel values, they will be useless, 
            /// when the new ZoomWindowForm is created, and must therefore be saved as horizontal - and 
            /// vertical axes values.
            /// </summary>
            public class ZoomRectangleCoordinates
            {
                private double _horizontalAxesMinimum = double.NaN;
                public double HorizontalAxesMinimum
                {
                    get
                    {
                        return _horizontalAxesMinimum;
                    }
                    set
                    {
                        _horizontalAxesMinimum = Ellab.Utilities.Convert.DateTimeConvertion.RoundOADate(value) ?? double.NaN;
                    }
                }

                private double _horizontalAxesMaximum = double.NaN;
                public double HorizontalAxesMaximum
                {
                    get
                    {
                        return _horizontalAxesMaximum;
                    }
                    set
                    {
                        _horizontalAxesMaximum = Ellab.Utilities.Convert.DateTimeConvertion.RoundOADate(value) ?? double.NaN;
                    }
                }

                public double VerticalAxesMinimum { get; set; }
                public double VerticalAxesMaximum { get; set; }

                public ZoomRectangleCoordinates()
                { 
                }

                public ZoomRectangleCoordinates(ZoomRectangleCoordinates zoomRectangle)
                {
                    HorizontalAxesMaximum = zoomRectangle.HorizontalAxesMaximum;
                    HorizontalAxesMinimum = zoomRectangle.HorizontalAxesMinimum;

                    VerticalAxesMinimum = zoomRectangle.VerticalAxesMinimum;
                    VerticalAxesMaximum = zoomRectangle.VerticalAxesMaximum;
                }
                
                public Rectangle GetZoomRectAngle(Chart parent)
                {
                    int x = parent.HorizontalAxis.ValueToPixelCoordinate(HorizontalAxesMinimum);
                    int y = parent.VerticalAxes[0].ValueToPixelCoordinate(VerticalAxesMaximum);
                    int width = parent.HorizontalAxis.ValueToPixelCoordinate(HorizontalAxesMaximum) - parent.HorizontalAxis.ValueToPixelCoordinate(HorizontalAxesMinimum);
                    int height = parent.VerticalAxes[0].ValueToPixelCoordinate(VerticalAxesMinimum) - parent.VerticalAxes[0].ValueToPixelCoordinate(VerticalAxesMaximum);

                    Rectangle zoomRectangle = new Rectangle(x, y, width, height);
                    //double zoomRectangleHeightWidthRatio = (double)zoomRectangle.Height / (double)zoomRectangle.Width;

                    //Rectangle chartBounds = parent.ChartBounds;
                    //double chartBoundsHeightWidthRatio = (double)chartBounds.Height / (double)chartBounds.Width;

                    //if (zoomRectangleHeightWidthRatio > chartBoundsHeightWidthRatio)
                    //{
                    //    double temp = System.Math.Ceiling(((double)zoomRectangle.Height) / chartBoundsHeightWidthRatio);
                    //    zoomRectangle.Width = System.Convert.ToInt32(temp);
                    //}

                    return zoomRectangle;
                }

                public void SetZoomRectangle(Chart parent)
                {
                    HorizontalAxesMinimum = parent.horizontalAxis.PixelCoordinateToValue(parent.TChart.Zoom.x0, int.MinValue); //For horizontal axes the y value is not used in this calculation
                    HorizontalAxesMaximum = parent.horizontalAxis.PixelCoordinateToValue(parent.TChart.Zoom.x1, int.MinValue);

                    VerticalAxesMinimum = parent.VerticalAxes[0].PixelCoordinateToValue(int.MinValue, parent.TChart.Zoom.y1); //For vertical axes the x value is not used in this calculation
                    VerticalAxesMaximum = parent.VerticalAxes[0].PixelCoordinateToValue(int.MinValue, parent.TChart.Zoom.y0);
                }
            }

            private ZoomRectangleCoordinates _zoomCoordinates = null;
            /// <summary>
            /// Axis zoom 'coordinates' values 
            /// </summary>
            public ZoomRectangleCoordinates ZoomCoordinates
            {
                get
                {
                    if (_zoomCoordinates == null)
                    {
                        _zoomCoordinates = new ZoomRectangleCoordinates();
                    }

                    return _zoomCoordinates;
                }
                private set
                {
                    _zoomCoordinates = value;
                }
            }

            /// <summary>
            /// Returns axis zoom 'coordinates' values
            /// </summary>
            /// <returns></returns>
            public ZoomRectangleCoordinates GetZoomCoordinates()
            {
                ZoomRectangleCoordinates zoomRectangle = new ZoomRectangleCoordinates();
                zoomRectangle.SetZoomRectangle(parent);
                return zoomRectangle;
            }

            /// <summary>
            /// Use this method to save the zoom pixel coordinates converted as axis values.
            /// </summary>
            public void SaveZoomCoordinates()
            {
                ZoomCoordinates.HorizontalAxesMinimum = parent.horizontalAxis.PixelCoordinateToValue(parent.TChart.Zoom.x1, parent.TChart.Zoom.y1);
                ZoomCoordinates.HorizontalAxesMaximum = parent.horizontalAxis.PixelCoordinateToValue(parent.TChart.Zoom.x0, parent.TChart.Zoom.y0);

                ZoomCoordinates.VerticalAxesMinimum = parent.VerticalAxes[0].PixelCoordinateToValue(parent.TChart.Zoom.x1, parent.TChart.Zoom.y1);
                ZoomCoordinates.VerticalAxesMaximum = parent.VerticalAxes[0].PixelCoordinateToValue(parent.TChart.Zoom.x0, parent.TChart.Zoom.y0);
            }

            /// <summary>
            /// Sets the axis zoom 'coordinates' values
            /// </summary>
            /// <param name="zoomRectangle"></param>
            public void SetZoomCoordinates(ZoomRectangleCoordinates zoomRectangle)
            {
                ZoomCoordinates.HorizontalAxesMaximum = zoomRectangle.HorizontalAxesMaximum;
                ZoomCoordinates.HorizontalAxesMinimum = zoomRectangle.HorizontalAxesMinimum;

                ZoomCoordinates.VerticalAxesMinimum = zoomRectangle.VerticalAxesMinimum;
                ZoomCoordinates.VerticalAxesMaximum = zoomRectangle.VerticalAxesMaximum;
            }

            /// <summary>
            /// By invoking CustomZoomInInternal, based upon earlier defined zoom coordinates, this will do a 'forced zoom'
            /// </summary>
            /// <param name="sender"></param>
            /// <param name="e"></param>
            private void CustomZoomIn(object sender, EventArgs e)
            {
                if (customZoomInDelegate == null)
                {
                    customZoomInDelegate = new Action<object, EventArgs>(CustomZoomInInternal);
                }

                parent.InvokeDelegate(customZoomInDelegate, sender, e);
            }
            private Action<object, EventArgs> customZoomInDelegate = null;
            /// <summary>
            /// Based upon earlier defined zoom coordinates, this will do a 'forced zoom'
            /// </summary>
            /// <param name="sender"></param>
            /// <param name="e"></param>
            private void CustomZoomInInternal(object sender, EventArgs e)
            {
                parent.TChart.Zoom.ZoomRect(ZoomCoordinates.GetZoomRectAngle(parent));
                //This 'forced zoom' will only take effect after TeeChart has been refreshed.
                parent.TChart.Refresh();
            }

            #endregion Custom Zooming

            private void ZoomOut(object sender, EventArgs e)
            {
                if (zoomHistory.Count == 0)
                    return;

                if (zoomOutDelegate == null)
                    zoomOutDelegate = new Action<object, EventArgs>(ZoomOutInternal);
                parent.InvokeDelegate(zoomOutDelegate, sender, e);
            }
            private Action<object, EventArgs> zoomOutDelegate = null;
            private void ZoomOutInternal(object sender, EventArgs e)
            {
                if (zoomHistory.Count == 0)
                    return;

                Dictionary<string, AxisZoomState> state = zoomHistory.Pop();
                foreach (string key in state.Keys)
                {
                    if (parent.axes.Contains(key))
                    {
                        parent.axes[key].Minimum = state[key].Minimum;
                        parent.axes[key].Maximum = state[key].Maximum;
                        parent.axes[key].AutomaticMinimum = state[key].AutomaticMinimum;
                        parent.axes[key].AutomaticMaximum = state[key].AutomaticMaximum;
                    }
                }
                parent.horizontalAxis.Minimum = state[HORIZONTAL_AXIS_NAME].Minimum;
                parent.horizontalAxis.Maximum = state[HORIZONTAL_AXIS_NAME].Maximum;
                parent.horizontalAxis.AutomaticMinimum = state[HORIZONTAL_AXIS_NAME].AutomaticMinimum;
                parent.horizontalAxis.AutomaticMaximum = state[HORIZONTAL_AXIS_NAME].AutomaticMaximum;
            }
        }
        private ChartZoom zoom = null;
        public ChartZoom Zoom
        {
            get
            {
                if (zoom == null)
                    zoom = new ChartZoom(this);
                return zoom;
            }
        }

        #endregion Zoom

        #region Delegates (Thread Safety)
        internal void InvokeDelegate(Delegate d, params object[] args)
        {
            if (this.InvokeRequired)
            {
                this.Invoke(d, args);
            }
            else
            {
                d.DynamicInvoke(args);
            }
        }
        #endregion

    }

    #region Helper Classes

    public class AxisCollection : KeyedCollection<string, Axis>
    {
        public AxisCollection() : base(StringComparer.InvariantCultureIgnoreCase) { }

        protected override string GetKeyForItem(Axis item)
        {
            return item.Unit;
        }
    }

    public class SeriesCollection : KeyedCollection<string, ISeries>
    {
        public SeriesCollection() : base(StringComparer.InvariantCultureIgnoreCase) { }

        protected override string GetKeyForItem(ISeries item)
        {
            return item.Name;
        }
    }

    public class ToolCollection : KeyedCollection<string, ITool>
    {
        public ToolCollection() : base(StringComparer.InvariantCultureIgnoreCase) { }

        protected override string GetKeyForItem(ITool item)
        {
            return item.Name;
        }
    }

    public class SeriesEventArgs : EventArgs
    {
        public ISeries Series
        {
            get;
            set;
        }

        public SeriesEventArgs(ISeries series)
        {
            Series = series;
        }
    }

    public class ToolEventArgs : EventArgs
    {
        public ITool Tool
        {
            get;
            set;
        }

        public ToolEventArgs(ITool tool)
        {
            Tool = tool;
        }
    }

    public class ColorLineEventArgs : EventArgs
    {
        public double Value { get; set; }

        public ColorLineEventArgs(double value)
        {
            Value = value;
        }
    }

    #endregion
}