﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Ellab.Charting
{
    public enum MarksTipOnMouseAction
    {
        Move,
        Click
    }

    public enum MarksTipStyle
    {
        XY
    }

    /// <summary>
    /// A TeeChart MarksTip tool is equivalent to Windows ToolTip. Assinging each MarksTip tool to a 
    /// serie, will insure that each serie is able to display specific relevant text in the ToolTip, by
    /// subscribing to MarksTipTextEvent event, and setting the text that should be displayed in the
    /// 'public static string MarksTipText' property.
    /// </summary>
    public class MarksTip : ITool, IDisposable
    {
        private Steema.TeeChart.Tools.MarksTip tMarksTip;
        private Steema.TeeChart.Tools.MarksTipGetTextEventHandler _steemaTeeChartToolsMarksTipGetTextEventHandler = null;

        #region Properties

        public Chart Chart
        {
            get;
            private set;
        }

        /// <summary>
        /// Convert ValSuite visible MarksTipOnMouseAction to Steema.TeeChart.Tools.MarksTipMouseAction
        /// and hence hides TeeChart from Valsuite.
        /// Can return null, but will throw an ArgumentNullException() if set to null.
        /// </summary>
        public MarksTipOnMouseAction? OnMouseAction
        {
            get
            {
                if (tMarksTip.MouseAction == Steema.TeeChart.Tools.MarksTipMouseAction.Click)
                    return MarksTipOnMouseAction.Click;
                else if (tMarksTip.MouseAction == Steema.TeeChart.Tools.MarksTipMouseAction.Move)
                    return MarksTipOnMouseAction.Move;
                else
                    return null;
            }
            set
            {
                if (value == MarksTipOnMouseAction.Click)
                    tMarksTip.MouseAction = Steema.TeeChart.Tools.MarksTipMouseAction.Click;
                else if (value == MarksTipOnMouseAction.Move)
                    tMarksTip.MouseAction = Steema.TeeChart.Tools.MarksTipMouseAction.Move;
                else
                    throw new ArgumentNullException();
            }
        }

        /// <summary>
        /// Convert ValSuite visible MarksTipStyle to Steema.TeeChart.Styles.MarksStyles
        /// and hence hides TeeChart from Valsuite.
        /// Not all posible combination impelmentated.
        /// Can return null, but will throw an ArgumentNullException() if set to null.
        /// </summary>
        public MarksTipStyle? Style
        {
            get
            {
                if (tMarksTip.Style == Steema.TeeChart.Styles.MarksStyles.XY)
                    return MarksTipStyle.XY;
                else
                    return null;
            }
            set
            {
                if (value == MarksTipStyle.XY)
                    tMarksTip.Style = Steema.TeeChart.Styles.MarksStyles.XY;
                else
                    throw new ArgumentNullException();
            }
        }

        /// <summary>
        /// Default 0 ms.
        /// </summary>
        public int MouseDelay
        {
            get { return tMarksTip.MouseDelay; }
            set { tMarksTip.MouseDelay = value; }
        }

        /// <summary>
        /// Default 10000 ms.
        /// </summary>
        public int MarksTipShowTime
        {
            get { return tMarksTip.HideDelay; }
            set { tMarksTip.HideDelay = value; }
        }

        private ISeries _serie;
        /// <summary>
        /// The Serie to whitch this MarksTip tool is to be assigned
        /// </summary>
        public ISeries Serie
        {
            get
            {
                if (_serie == null)
                {
                    Ellab.Debug.GlobalDebug.debugForm.WriteText("Charting.MarksTip: Serie used before it has been set", Debug.DebugType.TeeChart);
                }
                return _serie;
            }
            set
            {
                tMarksTip.Series = value.baseSeries;
                _serie = value;

                if (_steemaTeeChartToolsMarksTipGetTextEventHandler == null)
                {
                    _steemaTeeChartToolsMarksTipGetTextEventHandler = new Steema.TeeChart.Tools.MarksTipGetTextEventHandler(tMarksTip_GetText);
                    tMarksTip.GetText += _steemaTeeChartToolsMarksTipGetTextEventHandler;
                }
            }
        }

        /// <summary>
        /// The text, that should be showen.
        /// </summary>
        public static string MarksTipText { get; set; }

        public bool Active
        {
            get
            {
                return tMarksTip.Active;
            }
            set
            {
                tMarksTip.Active = value;
            }
        }

        #endregion Properties

        private MarksTip(Chart chart, string name, ISeries serie)
        {
            Chart = chart;
            Name = name;
            tMarksTip = new Steema.TeeChart.Tools.MarksTip(chart.TChart.Chart);

            tMarksTip.MouseAction = Steema.TeeChart.Tools.MarksTipMouseAction.Move;
            tMarksTip.MouseDelay = 0;
            tMarksTip.HideDelay = 10000;
            tMarksTip.Style = Steema.TeeChart.Styles.MarksStyles.XY;

            if (serie != null)
            {
                Serie = serie;
            }
        }

        /// <summary>
        /// While creating a MarksTip tool will set most of the default values, it is not assined to a specific series,
        /// thus will operate on all series in the given chart. Other MarksTip tools might have been assined to a serie,
        /// potentially confusing things up. Therefor when usign this constructor, remember to assign a serie to this MarksTip 
        /// tool, using property 'public ISeries Serie'.
        /// </summary>
        /// <param name="chart">Current Chart</param>
        /// <param name="name">Name of this instance of MarksTip tool</param>
        /// <returns>A new instance of a MarksTip object</returns>
        public static MarksTip Create(Chart chart, string name)
        {
            return new MarksTip(chart, name, null);
        }

        /// <summary>
        /// Creating a MarksTip tool, setting default values, and assineding it to a specific series, thus will operate on only 
        /// one serie in the given chart. 
        /// </summary>
        /// <param name="chart">Current Chart</param>
        /// <param name="name">Name of this instance of MarksTip tool</param>
        /// <param name="serie">The Serie to whitch this MarksTip tool is to be assigned</param>
        /// <returns>A new instance of a MarksTip object</returns>
        public static MarksTip Create(Chart chart, string name, ISeries serie)
        {
            return new MarksTip(chart, name, serie);
        }

        #region Events

        /// <summary>
        /// Subscribe to this event to set the text, that should be showen in the tooltip.
        /// The event handler of this event should set the MarksTipText property, as this is the 
        /// text string showen.
        /// Axis coordinates for this paticually series is given in the 
        /// MarksTipGetTextEventArgs 
        /// as a text string, and can be obtained something like this
        /// string[] text = e.Text.Split(' ');
        /// text[0] will hold the X value (eg. Date/time)
        /// text[1] the Y value (eg. Temperatur)
        /// </summary>
        public event EventHandler<MarksTipGetTextEventArgs> MarksTipTextEvent;

        private void tMarksTip_GetText(Steema.TeeChart.Tools.MarksTip sender, Steema.TeeChart.Tools.MarksTipGetTextEventArgs e)
        {
            try
            {
                if (MarksTipTextEvent != null)
                {
                    MarksTipGetTextEventArgs marksTipTextArgs = new MarksTipGetTextEventArgs(e.Text);
                    MarksTipTextEvent(this, marksTipTextArgs);

                    if (MarksTipText != null)
                    {
                        e.Text = MarksTipText;
                    }
                }
            }
            catch (Exception ex)
            {
                Ellab.Debug.GlobalDebug.debugForm.WriteText(ex);
            }
        }

        #endregion Events

        #region ITool Members

        public void Remove()
        {
            if (removeDelegate == null)
                removeDelegate = new Action(RemoveInternal);
            Chart.InvokeDelegate(removeDelegate);
        }
        private Action removeDelegate = null;
        private void RemoveInternal()
        {
            Chart.TChart.Tools.Remove(tMarksTip);
            Chart.RemoveTool(this);
        }

        public ToolType Type { get { return ToolType.MarksTip; } }

        public string Name { get; set; }

        public ITool CreateMirror(Chart chart)
        {
            MarksTip marksTip = new MarksTip(chart, this.Name, this.Serie);

            return marksTip;
        }

        #endregion

        #region IDisposable Members

        public void Dispose()
        {
            Remove();

            if (_steemaTeeChartToolsMarksTipGetTextEventHandler != null)
            {
                tMarksTip.GetText -= _steemaTeeChartToolsMarksTipGetTextEventHandler;
            }

            MarksTipTextEvent = null;
        }

        #endregion

        /// <summary>
        /// DO NOT USE
        /// </summary>
        private MarksTip()
        {
            throw new NotImplementedException();
        }

        #region Helper Classes

        public class MarksTipGetTextEventArgs : EventArgs
        {
            public string Text { get; set; }

            public MarksTipGetTextEventArgs(string text)
            {
                Text = text;
            }
        }

        #endregion Helper Classes
    }
}
