﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;

namespace Ellab.Charting
{
    public class NearestPoint : ITool, IDisposable
    {
        private Steema.TeeChart.Tools.NearestPoint tNearestPoint;
        private EventHandler _steemaTeeChartToolsNearestPointEventHandler = null;

        #region Properties

        public Chart Chart
        {
            get;
            private set;
        }

        private ISeries _serie;
        /// <summary>
        /// The Serie to whitch this NearestPoint tool is to be assigned
        /// </summary>
        public ISeries Serie
        {
            get
            {
                if (_serie == null)
                {
                    Ellab.Debug.GlobalDebug.debugForm.WriteText("Charting.NearestPoint: Serie used before it has been set", Debug.DebugType.TeeChart);
                }
                return _serie;
            }
            set
            {
                tNearestPoint.Series = value.baseSeries;
                _serie = value;
            }
        }

        public bool Active
        {
            get
            {
                return tNearestPoint.Active;
            }
            set
            {
                tNearestPoint.Active = value;
            }
        }

        #endregion

        private NearestPoint(Chart chart, string name, ISeries serie)
        {
            Chart = chart;
            Name = name;
            tNearestPoint = new Steema.TeeChart.Tools.NearestPoint(chart.TChart.Chart);

            if (serie != null)
            {
                Serie = serie;
            }

            tNearestPoint.Pen.Color = Color.Blue;
            tNearestPoint.Size = 5;
            tNearestPoint.Style = Steema.TeeChart.Tools.NearestPointStyles.Circle;

            tNearestPoint.Active = false;

            if (_steemaTeeChartToolsNearestPointEventHandler == null)
            {
                _steemaTeeChartToolsNearestPointEventHandler = new EventHandler(tNearestPoint_Change);
            }
            tNearestPoint.Change += _steemaTeeChartToolsNearestPointEventHandler;
        }

        public event EventHandler<NearestPointChangeEventArgs> NearestPointChangeDelegate = null;

        private void tNearestPoint_Change(object sender, EventArgs e)
        {
            try
            {
                if (NearestPointChangeDelegate != null)
                {
                    if (tNearestPoint.Point != -1)
                    {
                        NearestPointChangeEventArgs nearestPointChangeEventArgs = new NearestPointChangeEventArgs(
                            tNearestPoint.Series.XValues[tNearestPoint.Point],
                            tNearestPoint.Series.YValues[tNearestPoint.Point]);

                        NearestPointChangeDelegate(this, nearestPointChangeEventArgs);
                    }
                }
            }
            catch (Exception ex)
            {
                Ellab.Debug.GlobalDebug.debugForm.WriteText(ex);
            }
        }

        /// <summary>
        /// Creating a NearestPoint tool without setting the ISeries, will make it operate on all
        /// series in the given Chart.
        /// </summary>
        /// <param name="chart">Current Chart</param>
        /// <param name="name">Name of this instance of NearestPoint tool</param>
        /// <returns>A new instance of a NearestPoint object</returns>
        public static NearestPoint Create(Chart chart, string name)
        {
            return new NearestPoint(chart, name, null);
        }

        /// <summary>
        /// Creating a NearestPoint tool, setting default values, and assineding it to a specific series, thus will operate on only 
        /// one serie in the given chart. 
        /// </summary>
        /// <param name="chart">Current Chart</param>
        /// <param name="name">Name of this instance of NearestPoint tool</param>
        /// <param name="serie">The Serie to whitch this NearestPoint tool is to be assigned</param>
        /// <returns>A new instance of a NearestPoint object</returns>
        public static NearestPoint Create(Chart chart, string name, ISeries serie)
        {
            return new NearestPoint(chart, name, serie);
        }

        #region ITool Members

        public void Remove()
        {
            if (removeDelegate == null)
                removeDelegate = new Action(RemoveInternal);
            Chart.InvokeDelegate(removeDelegate);
        }
        private Action removeDelegate = null;
        private void RemoveInternal()
        {
            Chart.TChart.Tools.Remove(tNearestPoint);
            Chart.RemoveTool(this);
        }


        public ToolType Type { get { return ToolType.NearestPoint; } }

        public string Name { get; set; }

        public ITool CreateMirror(Chart chart)
        {
            throw new NotImplementedException();
        }

        #endregion

        #region IDisposable Members

        public void Dispose()
        {
            if (_steemaTeeChartToolsNearestPointEventHandler != null)
            {
                tNearestPoint.Change -= _steemaTeeChartToolsNearestPointEventHandler;
            }
            _steemaTeeChartToolsNearestPointEventHandler = null;
        }

        /// <summary>
        /// DO NOT USE
        /// </summary>
        private NearestPoint()
        {
            throw new NotImplementedException();
        }

        #endregion

        #region HelperClass
        public class NearestPointChangeEventArgs : EventArgs
        {
            public double X = double.NaN;
            public double Y = double.NaN;

            public NearestPointChangeEventArgs(double xValue, double yValue)
            {
                X = xValue;
                Y = yValue;
            }
        }
        #endregion
    }
}
