﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;

namespace Ellab.Charting
{
    public abstract class SeriesBase : ISeries
    {
        private Chart chart;
        protected Steema.TeeChart.Styles.Series tSeries;

        public Steema.TeeChart.Styles.Series baseSeries
        {
            get
            {
                return tSeries;
            }
        }

        public event EventHandler PointAdded;
        public event EventHandler PointRemoved;

        public Chart Chart
        {
            get { return chart; }
            protected set
            {
                chart = value;
                if (chart != null)
                {
                    tSeries.Chart = chart.TChart.Chart;
                    tSeries.CustomHorizAxis = chart.HorizontalAxis.TAxis;
                    tSeries.CustomVertAxis = chart.GetAxis(Unit).TAxis;
                }
            }
        }

        private string unit;
        public string Unit
        {
            get { return unit; }
            set
            {
                if (value == null || value == "")
                    value = "(Unknown Unit)";
                unit = value;
                if (Chart != null)
                {
                    tSeries.CustomVertAxis = Chart.GetAxis(unit).TAxis;
                    Chart.RemoveUnusedAxes();
                }
            }
        }

        private string title;
        public string Title
        {
            get { return title; }
            set
            {
                if (value != null)
                {
                    title = value;
                }
                else
                {
                    title = Unit;
                    Ellab.Debug.GlobalDebug.debugForm.WriteText("Axes Unit used as Title. Unit is: " + Unit, Debug.DebugType.TeeChart);
                }
            }
        }

        public SeriesType Type { get { return SeriesType.LineSeries; } }

        public string Name
        {
            get { return tSeries.Title; }
            set
            {
                if (nameDelegate == null)
                    nameDelegate = new Action<string>(SetName);
                chart.InvokeDelegate(nameDelegate, value);
            }
        }
        private Action<string> nameDelegate = null;
        private void SetName(string name)
        {
            tSeries.Title = name;
        }

        public bool Visible
        {
            get { return tSeries.Visible; }
            set
            {
                if (visibleDelegate == null)
                    visibleDelegate = new Action<bool>(SetVisible);
                Chart.InvokeDelegate(visibleDelegate, value);
            }
        }
        private Action<bool> visibleDelegate = null;
        private void SetVisible(bool value)
        {
            tSeries.Visible = value;
        }

        public System.Drawing.Color Color
        {
            get { return tSeries.Color; }
            set
            {
                if (colorDelegate == null)
                    colorDelegate = new Action<System.Drawing.Color>(SetColor);
                chart.InvokeDelegate(colorDelegate, value);
            }
        }
        private Action<System.Drawing.Color> colorDelegate = null;
        private void SetColor(System.Drawing.Color color)
        {
            tSeries.Color = color;
        }

        /// <summary>
        /// DO NOT USE. Use MarksTip instead
        /// </summary>
        private Charting.MarksTip _marksTip = null;
        public Charting.MarksTip MarksTip
        {
            get
            {
                if (_marksTip == null)
                {
                    _marksTip = MarksTip.Create(Chart, "MarksTip Tool", this);
                }
                return _marksTip;
            }
            set
            {
                _marksTip = value;
            }
        }

        /// <summary>
        /// DO NOT USE. Use NearestPoint instead
        /// </summary>
        private Charting.NearestPoint _nearestPoint = null;
        public Charting.NearestPoint NearestPoint
        {
            get
            {
                if (_marksTip == null)
                {
                    _marksTip = MarksTip.Create(Chart, "NearestPoint Tool", this);
                }
                return _nearestPoint;
            }
            set
            {
                _nearestPoint = value;
            }
        }

        public bool XValuesAsDataTime
        {
            set
            {
                tSeries.XValues.DateTime = value;
            }
        }

        public int Precision
        {
            get;
            set;
        }

        #region Data

        public class SeriesData : ISeriesData
        {
            private SeriesBase parent;
            internal SeriesData(SeriesBase _parent)
            {
                parent = _parent;
            }

            public int Count
            {
                get
                {
                    if (parent.tSeries.XValues.Count < parent.tSeries.YValues.Count)
                        return parent.tSeries.XValues.Count;
                    else
                        return parent.tSeries.YValues.Count;
                }
            }

            public Coordinate this[int index]
            {
                get
                {
                    if (index >= 0 && index < parent.tSeries.XValues.Count)
                        return new Coordinate(parent.tSeries.XValues[index], parent.tSeries.YValues[index]);
                    else
                        throw new IndexOutOfRangeException();
                }
            }

            public void SetX(int index, double xValue)
            {
                if (index < 0 || index >= parent.tSeries.XValues.Count)
                    throw new IndexOutOfRangeException();

                if (setXDelegate == null)
                    setXDelegate = new Action<int, double>(SetXInternal);
                parent.Chart.InvokeDelegate(setXDelegate, index, xValue);
            }
            public void SetY(int index, double yValue)
            {
                if (index < 0 || index >= parent.tSeries.YValues.Count)
                    throw new IndexOutOfRangeException();

                if (setYDelegate == null)
                    setYDelegate = new Action<int, double>(SetYInternal);
                parent.Chart.InvokeDelegate(setYDelegate, index, yValue);
            }

            /// <summary>
            /// Returns the Y-value at the given X-value.
            /// This will never attempt to interpolate a value! If the exact X-value given is not present, null is returned.
            /// </summary>
            /// <param name="xValue"></param>
            /// <returns></returns>
            public double? this[DateTime xValue]
            {
                get
                {
                    int? index = IndexOf(xValue);
                    if (index.HasValue)
                        return parent.tSeries.YValues[index.Value];
                    else
                        return null;
                }
                set
                {
                    if (!value.HasValue)
                        return;

                    int? index = IndexOf(xValue);
                    if (!index.HasValue)
                        return;

                    if (setYDelegate == null)
                        setYDelegate = new Action<int, double>(SetYInternal);
                    parent.Chart.InvokeDelegate(setYDelegate, index.Value, value);
                }
            }

            private Action<int, double> setXDelegate = null;
            private void SetXInternal(int index, double xValue)
            {
                if (index >= 0 && parent.tSeries.XValues.Count > index)
                    parent.tSeries.XValues[index] = xValue;
            }
            private Action<int, double> setYDelegate = null;
            private void SetYInternal(int index, double yValue)
            {
                if (index >= 0 && parent.tSeries.YValues.Count > index)
                    parent.tSeries.YValues[index] = yValue;
            }

            public void Add(double x, double y)
            {
                Add(DateTime.FromOADate(x), y);
            }

            public void Add(DateTime x, double y)
            {
                if (addDelegate == null)
                {
                    addDelegate = new Action<DateTime, double>(AddInternal);
                }

                parent.Chart.InvokeDelegate(addDelegate, x, y);
            }

            public void Add(double[] x, double[] y)
            {
                //CKBN If ever units on x axis is anything else than DataTime use this:
                //if (addArrayDelegate == null)
                //{
                //    addArrayDelegate = new Action<double[], double[]>(AddInternal);
                //}

                //parent.Chart.InvokeDelegate(addArrayDelegate, x, y);
                //CKBN otherwise always convert x to DataTime:

                DateTime[] dateTimeArray = new DateTime[x.Length];

                for (int i = 0; i < x.Length; i++)
                {
                    dateTimeArray[i] = DateTime.FromOADate(x[i]);
                }

                if (addArrayDateTimeDelegate == null)
                {
                    addArrayDateTimeDelegate = new Action<DateTime[], double[]>(AddInternal);
                }

                parent.Chart.InvokeDelegate(addArrayDateTimeDelegate, dateTimeArray, y);
            }

            public void Add(DateTime[] x, double[] y)
            {
                if (addArrayDateTimeDelegate == null)
                {
                    addArrayDateTimeDelegate = new Action<DateTime[], double[]>(AddInternal);
                }

                parent.Chart.InvokeDelegate(addArrayDateTimeDelegate, x, y);
            }

            private Action<DateTime, double> addDelegate = null;
            private void AddInternal(DateTime x, double y)
            {
                parent.tSeries.Add(x, y);
                SetHorizontalAxisLable();

                if (parent.PointAdded != null)
                {
                    parent.PointAdded(this, new EventArgs());
                }
            }

            private Action<double[], double[]> addArrayDelegate = null;
            private void AddInternal(double[] x, double[] y)
            {
                parent.tSeries.Add(x, y);
                SetHorizontalAxisLable();
            }

            private Action<DateTime[], double[]> addArrayDateTimeDelegate = null;
            private void AddInternal(DateTime[] x, double[] y)
            {
                parent.tSeries.Add(x, y);
                SetHorizontalAxisLable();
            }

            /// <summary>
            /// Will peform a check on whether the shown samples streches over a period grather than 
            /// 24 hours. If so, will setup TeeChart to show both the date and time.
            /// If samles streches over a period grather than 5 days, TeeChart will be setup to show
            /// only dates
            /// </summary>
            private void SetHorizontalAxisLable()
            {
                DateTime minDateTime = DateTime.FromOADate(parent.HorizontalAxis.Minimum);
                DateTime maxDateTime = DateTime.FromOADate(parent.HorizontalAxis.Maximum);

                if (maxDateTime <= minDateTime.AddHours((double)24))
                {
                    parent.HorizontalAxis.LabelFormat =
                        "HH" +
                        System.Threading.Thread.CurrentThread.CurrentCulture.DateTimeFormat.TimeSeparator +
                        "mm" +
                        System.Threading.Thread.CurrentThread.CurrentCulture.DateTimeFormat.TimeSeparator +
                        "ss";
                }
                else if (maxDateTime > minDateTime.AddHours((double)24) && maxDateTime < minDateTime.AddDays((double)5))
                {
                    parent.HorizontalAxis.LabelFormat =
                        "MM" +
                        System.Threading.Thread.CurrentThread.CurrentCulture.DateTimeFormat.DateSeparator +
                        "dd HH";
                }
                else //maxDateTime >= minDateTime.AddDays(5)
                {
                    parent.HorizontalAxis.LabelFormat =
                        "yyyy" +
                        System.Threading.Thread.CurrentThread.CurrentCulture.DateTimeFormat.DateSeparator +
                        "MM" +
                        System.Threading.Thread.CurrentThread.CurrentCulture.DateTimeFormat.DateSeparator +
                        "dd";
                }

                #region Test region
                //if (maxDateTime <= minDateTime.AddMinutes(1))
                //{
                //    parent.HorizontalAxis.LabelFormat = "HH:mm:ss";
                //}
                //else if (maxDateTime > minDateTime.AddMinutes(1) && maxDateTime < minDateTime.AddMinutes(2))
                //{
                //    parent.HorizontalAxis.LabelFormat = "MM/dd HH";
                //}
                //else //maxDateTime >= minDateTime.AddMinutes(2)
                //{
                //    parent.HorizontalAxis.LabelFormat = "yyyy/MM/dd";
                //}
                #endregion
            }

            public void Remove(DateTime xValue)
            {
                int? index = IndexOf(xValue);
                if (index.HasValue)
                    Remove(index.Value);
            }
            public void Remove(int index)
            {
                if (removeDelegate == null)
                    removeDelegate = new Action<int>(RemoveInternal);
                parent.Chart.InvokeDelegate(removeDelegate, index);
            }
            private Action<int> removeDelegate = null;
            private void RemoveInternal(int index)
            {
                parent.tSeries.Delete(index);
                if (parent.PointRemoved != null)
                    parent.PointRemoved(this, new EventArgs());
            }

            public void Clear()
            {
                if (clearDelegate == null)
                    clearDelegate = new Action(ClearInternal);
                parent.Chart.InvokeDelegate(clearDelegate);
            }
            private Action clearDelegate = null;
            private void ClearInternal()
            {
                parent.tSeries.Clear();
            }

            /// <summary>
            /// Attempts to interpolate the Y-value at the given X-value.
            /// Interpolation must occur between two existing points and so cannot be done before the first or after the last X-value. Attempting to do so will return null.
            /// </summary>
            /// <param name="xValue"></param>
            /// <returns></returns>
            public double? InterpolateValue(DateTime xValue)
            {
                return InterpolateValue(xValue.ToOADate());
            }
            /// <summary>
            /// Attempts to interpolate the Y-value at the given X-value.
            /// Interpolation must occur between two existing points and so cannot be done before the first or after the last X-value. Attempting to do so will return null.
            /// </summary>
            /// <param name="xValue"></param>
            /// <returns></returns>
            public double? InterpolateValue(double xValue)
            {
                int indexAfter = -1;
                for (int i = 0; i < parent.tSeries.XValues.Count; i++)
                {
                    if (parent.tSeries.XValues[i] > xValue)
                    {
                        indexAfter = i;
                        break;
                    }
                }

                if (indexAfter <= 0 || indexAfter > parent.tSeries.YValues.Count)    //Don't attempt interpolation for values lower than the first X-value
                    return null;

                double dx = parent.tSeries.XValues[indexAfter] - parent.tSeries.XValues[indexAfter - 1];
                double dy = parent.tSeries.YValues[indexAfter] - parent.tSeries.YValues[indexAfter - 1];

                if (dx != 0.0)
                    return dy * (xValue - parent.tSeries.XValues[indexAfter - 1]) / dx + parent.tSeries.YValues[indexAfter - 1];
                else
                    return (parent.tSeries.YValues[indexAfter] + parent.tSeries.YValues[indexAfter - 1]) / 2;
            }

            public int? IndexOf(DateTime xValue)
            {
                return IndexOf(xValue.ToOADate());
            }
            public int? IndexOf(double xValue)
            {
                for (int i = 0; i < parent.tSeries.XValues.Count; i++)
                {
                    if (parent.tSeries.XValues[i] == xValue && parent.tSeries.YValues.Count > i)
                        return i;
                }
                return null;
            }

            public DateTime MinimumX
            {
                get
                {
                    return DateTime.FromOADate(parent.tSeries.XValues.Minimum);
                }
            }
            public DateTime MaximumX
            {
                get
                {
                    return DateTime.FromOADate(parent.tSeries.XValues.Maximum);
                }
            }
            public double MinimumY
            {
                get
                {
                    return parent.tSeries.YValues.Minimum;
                }
            }
            public double MaximumY
            {
                get
                {
                    return parent.tSeries.YValues.Maximum;
                }
            }
        }
        private SeriesData data = null;
        public ISeriesData Data
        {
            get
            {
                if (data == null)
                {
                    data = new SeriesData(this);
                }

                return data;
            }
        }

        #endregion Data

        public Axis HorizontalAxis
        {
            get { return Chart.HorizontalAxis; }
        }

        public Axis VerticalAxis
        {
            get { return Chart.GetAxis(Unit); }
            set
            {
                this.Unit = value.Unit;
                this.Title = value.Title;
            }
        }

        /// <summary>
        /// Returns whether the series travels through the given pixel-coordinate.
        /// </summary>
        /// <param name="p"></param>
        /// <returns></returns>
        public bool IntersectsWith(Point p)
        {
            return IntersectsWith(p.X, p.Y);
        }
        /// <summary>
        /// Returns whether the series travels through the given pixel-coordinate.
        /// </summary>
        /// <param name="p"></param>
        /// <returns></returns>
        public bool IntersectsWith(int x, int y)
        {
            return SectionIntersectingWith(x, y) != -1;
        }
        /// <summary>
        /// Returns the section (0-based index) of the series that intersects with the given pixel-coordinate, or -1 if there is no intersection.
        /// A section being the line immediately following a point (the line between points 0 and 1 is section 0).
        /// </summary>
        /// <param name="p"></param>
        /// <returns></returns>
        public int SectionIntersectingWith(Point p)
        {
            return SectionIntersectingWith(p.X, p.Y);
        }
        /// <summary>
        /// Returns the section (0-based index) of the series that intersects with the given pixel-coordinate, or -1 if there is no intersection.
        /// A section being the line immediately following a point (the line between points 0 and 1 is section 0).
        /// </summary>
        /// <param name="p"></param>
        /// <returns></returns>
        public int SectionIntersectingWith(int x, int y)
        {
            return tSeries.Clicked(x, y);
        }

        public void RefreshDataSource()
        {
            tSeries.CheckDataSource();
        }

        public void Remove()
        {
            if (removeDelegate == null)
                removeDelegate = new Action(RemoveInternal);
            chart.InvokeDelegate(removeDelegate);
        }
        private Action removeDelegate = null;
        private void RemoveInternal()
        {
            Chart.TChart.Series.Remove(tSeries);
            Chart.RemoveSeries(this);
            tSeries.XValues.DateTime = true; //CKBN New TeeChart
        }

        public void RemoveAll()
        {
            if (removeAllDelegate == null)
                removeAllDelegate = new Action(RemoveAllInternal);
            chart.InvokeDelegate(removeAllDelegate);
        }
        private Action removeAllDelegate = null;
        private void RemoveAllInternal()
        {
            Chart.TChart.Series.RemoveAllSeries();
            //Chart.RemoveSeries(this);
            //tSeries.XValues.DateTime = true; //CKBN New TeeChart
        }

        #region Abstract

        public abstract ISeries CreateMirror(Chart chart);

        #endregion Abstract
    }
}
